from __future__ import absolute_import, print_function, unicode_literals

# utils.py contains utility functions that should require few, if any, 3rd party python libraries.

import re

# regular expressions
re_hgvs_with_gene_name = re.compile('^\w+\.\d+(?P<gene_name>\(\w+\)):')


def strip_gene_name_from_hgvs_text(hgvs_text):
    """ If gene name is embedded in the hgvs_text string, strip it out and
    return the HGVS string without the gene name part.

    :param: hgvs_text (str)
    :return: hgvs_text with gene_name removed (if needed)
    :rtype: str
    """
    # see if the gene name is embedded in the string, e.g. NM_003331.4(TYK2):c.3318_3319insC
    match = re_hgvs_with_gene_name.match(hgvs_text)
    if match:
        gene_str = match.groupdict()['gene_name']
        return hgvs_text.replace(gene_str, '')
    else:
        return hgvs_text



from __future__ import absolute_import, unicode_literals


class MetaVariantError(Exception):
    """ Generic exception class from which all metavariant exceptions are subclassed. """


class RejectedSeqVar(MetaVariantError):
    """ Indicates a sequence variant that cannot be considered functionally
    useful for the purposes of lookup in databases (pubtator, etc).

    Message should contain information about why the SequenceVariant fails.
    """
    pass


class CriticalHgvsError(MetaVariantError):
    """ Raised when a SequenceVariant cannot be parsed from the input hgvs_text; should be used only in cases
        where failure to create a SequenceVariant means a complete (critical) failure to proceed with LVG.

        Message should contain information passed along from the HgvsParseError.

        e.g. "HGVSParseError(u'NM_004628.4:c.621_622ins83: char 24: Syntax error',)"
    """
    pass


class LOVDRemoteError(MetaVariantError):
    """ Raised when response from LOVD contains a not-ok status code.
    Error message should contain status code. 
    """
    pass





import os, socket
from configparser import ConfigParser

import hgvs.dataproviders.uta

PKGNAME = 'metavariant'

DEBUG = bool(os.getenv('%s_DEBUG' % PKGNAME, False))
ENV = os.getenv('%s_ENV' % PKGNAME, 'dev')

UTA_HOST = os.getenv('UTA_HOST', 'default')
UTA_PORT = os.getenv('UTA_PORT', 5432)
UTA_SCHEMA = os.getenv('UTA_SCHEMA', 'uta_20180821')
UTA_TIMEOUT = os.getenv('UTA_TIMEOUT', 3)
UTA_USER = os.getenv('UTA_USER', 'uta_admin')
UTA_PASS = os.getenv('UTA_PASS', 'uta_admin')

####
import logging
log = logging.getLogger(PKGNAME)
if DEBUG:
    log.setLevel(logging.DEBUG)
else:
    log.setLevel(logging.INFO)
####

#log.debug('%s config dir: %s' % (PKGNAME, CFGDIR))
#log.debug('%s env: %s' % (PKGNAME, ENV))

def get_uta_connection(host=UTA_HOST, port=UTA_PORT, timeout=UTA_TIMEOUT, schema=UTA_SCHEMA, 
                        username=UTA_USER, password=UTA_PASS):
    """ Returns an open connection to a UTA host at given coordinates, if possible.
    
    If host=='default', returns connection to default UTA server (uta.biocommons.org).

    :return: open UTA connection
    :raises: socket, uta, hgvs Exceptions
    """

    timeout = int(timeout)
    port = int(port)

    uta_cnxn_tmpl = 'postgresql://{user}:{pwd}@{host}:{port}/uta/{schema}/'

    if host == 'default':
        return hgvs.dataproviders.uta.connect()

    socket.create_connection((host, port), timeout=timeout)
    try:
        return hgvs.dataproviders.uta.connect(uta_cnxn_tmpl.format(host=host, port=port, schema=schema, user=username, pwd=password), pooling=True)
    except:
        from IPython import embed; embed()




from __future__ import print_function, unicode_literals

import logging
import re

from .config import PKGNAME
from .exceptions import RejectedSeqVar

log = logging.getLogger(PKGNAME)

amino_acid_map = { 'Ala': 'A',
                   'Arg': 'R',
                   'Asn': 'N',
                   'Asp': 'D',
                   'Cys': 'C',
                   'Glu': 'E',
                   'Gln': 'Q',
                   'Gly': 'G',
                   'His': 'H',
                   'Ile': 'I',
                   'Leu': 'L',
                   'Lys': 'K',
                   'Met': 'M',
                   'Phe': 'F',
                   'Pro': 'P',
                   'Ser': 'S',
                   'Thr': 'T',
                   'Trp': 'W',
                   'Tyr': 'Y',
                   'Val': 'V',
                 }

# ... https://pypi.python.org/pypi/bidict/ ...?
amino_acid_map_reverse = { 'A': 'Ala',
                           'R': 'Arg',
                           'N': 'Asn',
                           'D': 'Asp',
                           'C': 'Cys',
                           'E': 'Glu',
                           'Q': 'Gln',
                           'G': 'Gly',
                           'H': 'His',
                           'I': 'Ile',
                           'L': 'Leu',
                           'K': 'Lys',
                           'M': 'Met',
                           'F': 'Phe',
                           'P': 'Pro',
                           'S': 'Ser',
                           'T': 'Thr',
                           'W': 'Trp',
                           'Y': 'Tyr',
                           'V': 'Val',
                         }


dna_nucleotides = ['A','C','T','G']
#rna_nucleotides = ['A','C','U','G']

official_to_slang_map = {'>': ['->', '-->', '/'],
                         'Ter': ['*', 'X'],
                        }

re_aminochange_long = re.compile('(?P<aminochange>\(?[A-Za-z]{3}\d+[A-Za-z]{3}\)?)')
re_aminochange_short = re.compile('(?P<aminochange>[{short_as}]\d+[{short_as}]\)?)'.format(short_as=''.join(amino_acid_map.values())))

re_aminochange_comp_long = re.compile('\(?(?P<ref>[A-Za-z]{3})(?P<pos>\d+)(?P<alt>[A-Za-z]{3})\)?')
re_aminochange_comp_short = re.compile('(?P<ref>[{short_as}])(?P<pos>\d+)(?P<alt>[{short_as}])'.format(short_as=''.join(amino_acid_map.values())))


def findall_aminochanges_in_text(text):
    """ Returns a LIST of all strings that appear to be amino acid change descriptions,
    e.g.: ['(Cys344Tyr)', 'C344Y', 'Cys344Tyr']

    If amino change is in parenthesis, this function returns a string with parenthesis removed.

    :param text: (str)
    :return: (list)
    """
    longs = re_aminochange_long.findall(text)
    shorts = re_aminochange_short.findall(text)
    return longs + shorts

def parse_components_from_aminochange(aminochange):
    """ Returns a dictionary containing (if possible) 'ref', 'pos', and 'alt'
    characteristics of the supplied aminochange string.

    If aminochange does not parse, returns None.

    :param aminochange: (str) describing amino acid change
    :return: dict or None
    """
    match = re_aminochange_comp_long.match(aminochange)
    if match:
        # reverse long-form amino strings to short-form.
        stuff = match.groupdict()
        return {'ref': amino_acid_map[stuff['ref']],
                'pos': stuff['pos'],
                'alt': amino_acid_map[stuff['alt']],
               }
    else:
        match = re_aminochange_comp_short.match(aminochange)
        return match.groupdict()
    return None

def lowercase_all_the_keys(some_dict):
    """ lowercases all the keys in the supplied dictionary.

    :param: dict
    :return: dict
    """
    return dict((key.lower(), val) for key, val in some_dict.items())


class VariantComponents(object):
    """
    VariantComponents

    Takes an hgvs.SequenceVariant object and parses it into components (position, reference,
    alternate, edit type, and sequence type), allowing further lexical manipulation for the
    purposes of generating pubmed evidence source "slang".

    Can also be instantiated from keyword components, helping organize this information into 
    an object, which will also perform seqtype inference (if necessary) and official posedit
    syntax construction.

    Note: FrameShift (FS) and Duplication (DUP) are currently not well supported.

    Attributes:

        seqtype: the sequence type of this seqvar (one of 'c', 'g', 'g', 'n')
        edittype: the type of mutation represented by this variant ('SUB', 'DEL', 'FS', etc)
        pos: position of the edit
        ref: reference sequence at given position
        alt: alternate (or "wildtype") at given position

    Properties:

        posedit: returns the HGVS "official" construction of this seqvar's position + edit information.
        posedit_slang: returns a list of algorithmically generated "slang" for given seqvar's posedit.

    Usage:

        # from SequenceVariant object:
        comp = VariantComponents(seqvar)

        # from individual components:
        comp = VariantComponents(seqtype='c', edittype='SUB', pos='128', ref='C', alt='T')
    """

    def __init__(self, seqvar=None, aminochange='', **kwargs):
        kwargs = lowercase_all_the_keys(kwargs)

        self.seqvar = seqvar
        if self.seqvar:
            self.seqtype, self.edittype, self.ref, self.pos, self.alt = self.parse(seqvar)
            #TODO: get FS_Pos and DupX out of seqvar when applicable.
            self.fs_pos = ''
            self.dupx = ''

        elif aminochange:
            # parse Ref, Pos, Alt from string 
            aminodict = parse_components_from_aminochange(aminochange)
            if aminodict:
                self.__dict__.update(aminodict)
            else:
                raise RejectedSeqVar('Could not parse aminochange string "%s" into components' % aminochange)
            self.seqtype = 'p'
            self.edittype = ''
            self.fs_pos = ''
            self.dupx = ''

        else:
            # names of keywords match capitalization used in MySQL m2p_* tables in pubtatordb
            self.seqtype = kwargs.get('seqtype', '').strip()
            self.edittype = kwargs.get('edittype', '').strip()
            self.ref = kwargs.get('ref', '').strip()
            self.pos = kwargs.get('pos', '').strip()
            self.alt = kwargs.get('alt', '').strip()
            self.fs_pos = kwargs.get('fs_pos', '').strip()
            self.dupx = kwargs.get('dupx', '').strip()

        if self.edittype.upper() == 'DELINS':
            # normalize DELINS to INDEL (synonymous)
            self.edittype = 'INDEL'

        if not self.seqtype:
            self.seqtype = self._infer_seqtype()

    def _infer_seqtype(self):
        # If SeqType is none and REF in [u] or ALT in [u] --> then RNA
        # If SeqType is none and REF in [AminoAcidsList] and ALT in [AminoAcidsList] --> then Protein
        # If SeqType is none and REF in [a,c,t,g] and ALT in [a,c,t,g] --> then DNA or RNA
        
        refalt = self.ref.upper() + self.alt.upper()

        if 'u' in refalt:
            # Definitely RNA: there's no "U" amino acid and no "U" in the DNA nucleotides.
            return 'n'

        for char in refalt:
            if char not in dna_nucleotides:
                if char in list(amino_acid_map.values()):
                    return 'p'

        # it's "probably" DNA, but we can't know for sure.
        return ''

    @staticmethod
    def parse(seqvar):
        """ return tuple of sequence variant components as
        (seqtype, edittype, ref, pos, alt)
        """
        if seqvar.type.strip() == '':
            log.warn('SequenceVariant has empty seqtype. (%r)' % seqvar)
            seqtype = ''
        else:
            seqtype = seqvar.type

        ref = alt = edittype = pos = ''

        #if seqvar.posedit.edit in ['?', '=']:
        if hasattr(seqvar.posedit.edit, 'lower'):
            # if the edit attribute is a str or unicode type thing:
            raise RejectedSeqVar('SequenceVariant missing edit information. (%r)' % seqvar)

        edittype = seqvar.posedit.edit.type.upper()

        try:
            ref = seqvar.posedit.edit.ref
            alt = seqvar.posedit.edit.alt
        except AttributeError as error:
            # We expect a 'dup' variant not to have an alt. warn us if this is a different situation.
            if not seqvar.posedit.edit.type == 'dup':
                log.warn('SequenceVariant %s: %s', seqvar, error)

        # Termination (STOP) codon: normalize '*' to 'X'
        if alt == '*':
            alt = 'X'

        if seqtype == 'p':
            try:
                pos = '%s' % seqvar.posedit.pos.start.pos
                ref = '%s' % seqvar.posedit.pos.start.aa
            except AttributeError:
                raise RejectedSeqVar('Protein entry incomplete (unusable).')

        else:
            if seqvar.posedit.pos.end != seqvar.posedit.pos.start:
                # compose a "range"
                pos = '%s_%s' % (seqvar.posedit.pos.start, seqvar.posedit.pos.end)
            else:
                pos = '%s' % seqvar.posedit.pos.start

        return seqtype, edittype, ref, pos, alt

    def to_mysql_dict(self):
        outd = { 'Ref': self.ref,
                 'Alt': self.alt,
                 'SeqType': self.seqtype,
                 'EditType': self.edittype,
                 'Pos': self.pos,
                 }

        if self.edittype == 'FS':
            outd['FS_Pos'] = self.fs_pos

        if self.edittype == 'DUP':
            outd['DupX'] = self.dupx

        return outd

    @property
    def posedit(self):
        """ Returns the canonical lexeme representing this variant's position and edit information. 
        If this variant represents an amino acid change, the long-form names will be used.
        """
        if self.seqvar:
            return '%s' % self.seqvar.posedit
        elif self.seqtype == 'p':
            # i.e. if we instantiated with the aminochange string
            ref = amino_acid_map_reverse[self.ref]  
            alt = amino_acid_map_reverse[self.alt]  
            return '%s%s%s' % (ref, self.pos, alt)

    def _posedit_slang_protein(self):
        out = set()
        posedit = self.posedit.replace('(', '').replace(')', '')
        for item in official_to_slang_map['Ter']:
            out.add(posedit.replace('Ter', item))

        fs_pos = posedit.find('fs')
        if fs_pos > -1:
            out.add(posedit[:fs_pos + len('fs')])
        else:
            # e.g. Lys2569Gly produces "K2569G"
            out.add('%s%s%s' % (self.ref, self.pos, self.alt))
        return list(out)

    def _posedit_slang_SUB(self):
        """ Handles the Substitution case for generating posedit slang from Components. """
        # Examples based on input hgvs_text 'NM_014874.3:c.891C>T'
        out = set()

        # E.g. ["891C>T", "891C->T", "891C-->T", "891C/T"]
        for slang_symbol in official_to_slang_map['>']:
            out.add(self.posedit.replace('>', slang_symbol))

        # TODO: delins variation (c.114_115delinsA from c.114_115G>A)

        # E.g. "C891T"
        out.add(self.ref + self.pos + self.alt)
        return list(out)

    def _posedit_slang_DEL(self):
        """ Handles the Deletion case for generating posedit slang from Components. """
        # Examples based on input hgvs_text 'NM_007294.3:c.4964_4982delCTGGCCTGACCCCAGAAGA'
        #
        # E.g. '4964_4982del'
        return [self.pos + 'del']

    def _posedit_slang_DUP(self):
        """ Handles the Duplication case for generating posedit slang from Components. 
            
        Example:
            posedit c.6869dupA --> posedit_slang '689dup'
        """
        return [self.pos + 'dup']

    def _posedit_slang_INDEL(self):
        """ Handles the Insertion-Deletion (aka DELINS) case for generating posedit slang from Components.

        Example:
            posedit c.112_117delAGGTCAinsTG --> posedit_slang 'c.112_117delinsTG'
        """
        return []

    def _posedit_slang_INS(self):
        """ Handles the Insertion case for generating posedit slang from Components.

        Returns an empty list (no slang terms known).
        """
        return []

    @property
    def posedit_slang(self):
        """ If supported, returns alternative lexeme that may represent this variant's position and edit info in the wild. """
        if self.seqtype == 'p':
            return self._posedit_slang_protein()

        try:
            slang_method = getattr(self, '_posedit_slang_%s' % self.edittype)
            return slang_method()
        except AttributeError:
            raise NotImplementedError('Cannot currently handle EditType %s' % self.edittype)

    def to_dict(self):
        return self.__dict__

    def __str__(self):
        return '%r' % self.__dict__

    def __repr__(self):
        return '%r' % self.__dict__




from __future__ import absolute_import, print_function, unicode_literals

import re
import json
import logging

import hgvs.parser
import hgvs.assemblymapper 
from hgvs.exceptions import HGVSDataNotAvailableError, HGVSParseError

from .components import VariantComponents
from .config import get_uta_connection, PKGNAME
from .exceptions import CriticalHgvsError, RejectedSeqVar
from .utils import strip_gene_name_from_hgvs_text

log = logging.getLogger(PKGNAME)

# === UTA Connection setup. === #
uta = get_uta_connection()
mapper = hgvs.assemblymapper.AssemblyMapper(uta)
hgvs_parser = hgvs.parser.Parser()


def _seqvar_map_func(in_type, out_type):
    func_name = '%s_to_%s' % (in_type, out_type)
    return getattr(mapper, func_name)


def seqvar_length(seqvar):
    comp = VariantComponents(seqvar)
    return len(comp.posedit)


def variant_to_gene_name(seqvar):
    """
    Get HUGO Gene Name (Symbol) for given sequence variant object.

    Input seqvar must be of type 'n', 'c', or 'p'.

    :param variant: hgvs.SequenceVariant
    :return: string gene name (or None if not available).
    """
    if seqvar.type in ['n', 'c', 'p']:
        try:
            tx_identity = uta.get_tx_identity_info(seqvar.ac)
        except HGVSDataNotAvailableError:
            return None

        if tx_identity is not None:
            return tx_identity[-1]
        else:
            return None
    else:
        return None


def _seqvar_to_seqvar(seqvar, base_type, new_type, transcript=None, maxlen=None):
    """ Using UTA, translate the input seqvar (SequenceVariant object) into 
    the desired new_type of sequence variant.  If the base_type is 'g', a transcript
    label will be required.

    (Optional) Specify `maxlen` (int) characters to restrict resultant SequenceVariant 
    to a reasonable size. Returns None if str(SequenceVariant) > maxlen.

    :param seqvar: (SequenceVariant)
    :param base_type: (str) single-letter abbrev for variant type to map FROM
    :param new_type: (str) single-letter abbrev for variant type to map TO
    :param transcript: (str) [default: None]
    :param maxlen: (int) max length of resultant str(SequenceVariant) to allow
    :return: SequenceVariant or None
    """

    if base_type == new_type:
        return None

    if base_type == 'p':
        return None

    if new_type == 'p' and base_type == 'g':
        return None

    map_seqvar = _seqvar_map_func(base_type, new_type)

    result_seqvar = None

    try:
        if base_type == 'g':
            if transcript:
                result_seqvar = map_seqvar(seqvar, transcript)
            else:
                return None
        result_seqvar = map_seqvar(seqvar)
    except NotImplementedError:
        log.debug('Cannot map %s to %s: hgvs raised NotImplementedError', seqvar, new_type)
        return None
    except HGVSDataNotAvailableError as error:
        log.debug('Cannot map %s to %s: hgvs raised HGVSDataNotAvailableError (%r)', seqvar, new_type, error)
        return None
    except Exception as error:
        # catch the general case to be robust around the hgvs library's occasional volatility.
        log.debug('Cannot map %s to %s: unexpected Exception (%r)', seqvar, new_type, error)
        return None

    # if sequence variant maps out to longer than maxlen chars, return None
    try:
        if maxlen and seqvar_length(result_seqvar) > maxlen:
            return None
    except RejectedSeqVar:
        # generally, a "rejected" seqvar is a protein with an empty edit, which is OK to keep.
        pass

    return result_seqvar


class VariantLVG(object):

    #TODO: remove these two class variables (when ready)...
    VERSION = '0.0.3'
    LVG_MODE = 'lvg'

    def __init__(self, hgvs_text_or_seqvar, **kwargs):
        """ Creates a VariantLVG object given an HGVS string or SequenceVariant object.

        Enrichment:

            VariantLVG provides for "enrichment" of lexical variant generation by allowing
            more transcripts and variations to be supplied at instantiation. Just use the 
            appropriate keyword for the type of information, remembering that the "enrichment"
            keyword arguments are all lists.

        Example:

            VariantLVG('some_hgvs_text', gene_name='GWAR', hgvs_g=['relataed_hgvs_g'], transcripts=['19058.5', '101101.3'])

        Keywords:
            hgvs_c (list): see Enrichment above
            hgvs_g (list): ""
            hgvs_n (list): ""
            hgvs_p (list): ""
            gene_name: accepts any string as gene name (should be HGNC standardized)
            transcripts (list): list of strings describing valid alternative transcripts for seqvar
            seqvar_max_len (int): restrict posedit lengths to this number of characters (or fewer).    
        """

        self.hgvs_text = strip_gene_name_from_hgvs_text('%s' % hgvs_text_or_seqvar)
        self.seqvar = self.parse(hgvs_text_or_seqvar)

        self._gene_name = kwargs.get('gene_name', None)

        seqvar_max_len = kwargs.get('seqvar_max_len', None)


        if self.seqvar is None:
            raise CriticalHgvsError('Cannot create SequenceVariant from input %s (see hgvs_lexicon log)' % hgvs_text_or_seqvar)

        # initialize transcripts list
        self.transcripts = set(kwargs.get('transcripts', []))

        # fill in all the different ways to talk about this variant in each sequence type.
        self.variants = {'g': dict(), 'c': dict(), 'n': dict(), 'p': dict()}

        # collect any variants that were supplied at instantiation ("enrichment")
        for input_hgvs_c in kwargs.get('hgvs_c', []):
            self.variants['c'][str(input_hgvs_c)] = Variant(input_hgvs_c)

        for input_hgvs_g in kwargs.get('hgvs_g', []):
            self.variants['g'][str(input_hgvs_g)] = Variant(input_hgvs_g)

        for input_hgvs_n in kwargs.get('hgvs_n', []):
            self.variants['n'][str(input_hgvs_n)] = Variant(input_hgvs_n)

        for input_hgvs_p in kwargs.get('hgvs_p', []):
            self.variants['p'][str(input_hgvs_p)] = Variant(input_hgvs_p)

        try:
            self.variants[self.seqvar.type][str(self.seqvar)] = self.seqvar
        except KeyError:
            log.warn('Ignoring supplied SequenceVariant of type "%s" (not supported) -- (input was %s).' % (self.seqvar.type, self.seqvar))

        if self.variants['c']:
            # attempt to derive all 4 types of SequenceVariants from all available 'c'.
            for var_c in list(self.variants['c'].values()):
                for this_type, value in list(self.variants.items()):
                    new_seqvar = _seqvar_to_seqvar(var_c, 'c', this_type, maxlen=seqvar_max_len)
                    if new_seqvar:
                        self.variants[this_type][str(new_seqvar)] = new_seqvar

        # Now that we have a 'g', collect all available transcripts.
        if self.variants['g']:
            for var_g in list(self.variants['g'].values()):
                for trans in self.get_transcripts(var_g):
                    self.transcripts.add(trans)

        # In the case of starting with a 'g' type...
        if self.seqvar.type == 'g' and self.transcripts:
            # we still need to collect 'c' and 'n' variants
            for trans in self.transcripts:
                var_c = _seqvar_to_seqvar(self.seqvar, 'g', 'c', trans, maxlen=seqvar_max_len)
                if var_c:
                    self.variants['c'][str(var_c)] = var_c
                var_n = _seqvar_to_seqvar(self.seqvar, 'g', 'n', trans, maxlen=seqvar_max_len)
                if var_n:
                    self.variants['n'][str(var_n)] = var_n

        # With a list of transcripts, we can do g_to_c and g_to_n
        if self.transcripts:
            for trans in self.transcripts:
                for var_g in list(self.variants['g'].values()):
                    # Find all available 'c'
                    if not trans.startswith('NR'):
                        var_c = _seqvar_to_seqvar(var_g, 'g', 'c', trans, maxlen=seqvar_max_len)
                        if var_c:
                            self.variants['c'][str(var_c)] = var_c

                    # Find all available 'n'
                    var_n = _seqvar_to_seqvar(var_g, 'g', 'n', trans, maxlen=seqvar_max_len)
                    if var_n:
                        self.variants['n'][str(var_n)] = var_n

        # map all newly found 'c' to 'p'
        for var_c in list(self.variants['c'].values()):
            new_seqvar = _seqvar_to_seqvar(var_c, 'c', 'p', maxlen=seqvar_max_len)
            if new_seqvar:
                self.variants['p'][str(new_seqvar)] = new_seqvar

    @property
    def gene_name(self):
        """ Lazy-loaded gene name based on hgvs lookup. """
        if self._gene_name is None:
            # Aggregate all our SequenceVariant objects into a single list.
            seqvars = list(self.variants['c'].values())
            seqvars.extend(list(self.variants['n'].values()))
            seqvars.extend(list(self.variants['p'].values()))
            
            # try each seqvar; take the first gene name that appears, and stop there.
            for seqvar in seqvars:
                name = variant_to_gene_name(seqvar)
                if name:
                    self._gene_name = name
                    break
        return self._gene_name

    @staticmethod
    def get_transcripts(var_g):
        """Supply a Genomic variant (var_g) to find its related transcripts.
        
        Be Aware: this uses Assembly GRCh38.

        :param var_g: (str)
        :returns: list of transcripts associated with this variant.
        """
        #TODO: Allow passing 'assembly' as keyword to change assembly.
        return mapper.relevant_transcripts(var_g)

    @staticmethod
    def parse(hgvs_text_or_seqvar):
        """ Parse input through hgvs_parser if text, do nothing if SequenceVariant.
        Return SequenceVariant object.

        Allow all potential hgvs parsing errors and type errors to flow upwards.

        :param hgvs_text_or_seqvar: string or SequenceVariant object
        :return: SequenceVariant object
        """
        if type(hgvs_text_or_seqvar) == hgvs.sequencevariant.SequenceVariant:
            return hgvs_text_or_seqvar

        hgvs_text = strip_gene_name_from_hgvs_text(hgvs_text_or_seqvar)

        try:
            return hgvs_parser.parse_hgvs_variant(str(hgvs_text))
        except HGVSParseError as error:
            log.info('Cannot create SequenceVariant from hgvs_text "%s": %r', hgvs_text, error)
            # Examples:
            #  HGVSParseError(u'NP_068780.2:p.Tyr?His: char 17: expected a digit',)
            #  HGVSParseError(u'NM_004628.4:c.621_622ins83: char 24: Syntax error',)
            return None

    @property
    def hgvs_c(self):
        return list(self.variants['c'].keys())

    @property
    def hgvs_g(self):
        return list(self.variants['g'].keys())

    @property
    def hgvs_p(self):
        return list(self.variants['p'].keys())

    @property
    def hgvs_n(self):
        return list(self.variants['n'].keys())

    @property
    def seqvars(self):
        """ returns a flat list of all SequenceVariant objects contained in
        the self.variants dictionary """
        out = []
        for seqvar_dict in (self.variants.values()):
            out = out + list(seqvar_dict.values())
        return out

    def to_dict(self, with_gene_name=True):
        """Returns contents of object as a 2-level dictionary.

        Supply with_gene_name = True [default: True] to return gene_name as well.

        (gene_name is a lazy-loaded magic attribute, and may take a second or two).
        """
        outd = {'variants': {},
                'transcripts': list(self.transcripts),
                'seqvar': self.seqvar,
                'hgvs_text': self.hgvs_text,
               }

        if with_gene_name:
            outd['gene_name'] = self.gene_name

        # turn each seqvar dictionary into just a list of its values (the seqvar objects).
        for seqtype, seqvar_dict in (self.variants.items()):
            outd['variants'][seqtype] = list(seqvar_dict.values())

        return outd

    def _simple_dict(self):
        return {'gene_name': self.gene_name,
                'hgvs_c': self.hgvs_c,
                'hgvs_g': self.hgvs_g,
                'hgvs_n': self.hgvs_n,
                'hgvs_p': self.hgvs_p,
                'hgvs_text': self.hgvs_text,
                'transcripts': list(self.transcripts),
               }

    def to_json(self):
        """ Returns a JSON representation of this object, one that can be used to 
        instantiate this object again using the "from_json" class method.

        :return: (str) json repr
        """
        outd = self._simple_dict()
        return json.dumps(outd)

    @classmethod
    def from_json(cls, json_str):
        """ Allows instantiation of VariantLVG object from pre-established values 
        set in a JSON data structure.  The structure is as follows:
                
                {
                'hgvs_text': <hgvs_text>,
                'gene_name': <gene_name>,
                'hgvs_c': [hgvs_c1,hgvs_c2...],
                'hgvs_g': [hgvs_g1,hgvs_g2...],
                'hgvs_n': [hgvs_n1,hgvs_n2...],
                'hgvs_p': [hgvs_p1,hgvs_p2...],
                'transcripts': [<transcript1>,<transcript2>...]
               }

        At least hgvs_text must be set.
        """
        inpd = json.loads(json_str)
        hgvs_text = inpd.pop('hgvs_text')

        return cls(hgvs_text, **inpd)

    def __str__(self):
        out = 'HGVS input: %s\n' % self.hgvs_text
        out += '%r' % self.seqvar
        return out


### API Convenience Functions

Variant = VariantLVG.parse




from __future__ import absolute_import, print_function, unicode_literals

import sys, re

import requests
import xmltodict

from lxml import etree
from lxml.html import HTMLParser

from .exceptions import LOVDRemoteError

re_transcript = re.compile('\/variants\/\w+\/(?P<transcript>NM_\d+.\d+)\?')

LOVD_DEFAULT_DOMAIN = 'databases.lovd.nl'

LOVD_API_GENE_VARIANTS_URL = 'http://{domain}/shared/api/rest.php/variants/{symbol}'
LOVD_PAGE_GENE_VARIANTS_URL = 'http://{domain}/shared/view/{symbol}#page_size=1000&page=1'

re_pubmed_id = re.compile('pubmed\/(?P<pmid>\d+)')
re_doi = re.compile(r'(10[.][0-9]{2,}(?:[.][0-9]+)*/(?:(?!["&\'])\S)+)')


def construct_hgvs_name(transcript, coding_sequence):
    return '%s:%s' % (transcript, coding_sequence)


class LOVDVariant(object):

    def __init__(self, vdict=None, table_row=None, transcript=None, has_haplotype=False):

        self.gene_name = None
        self.title = None
        self.transcript = transcript
        self.hgvs_text = None
        self.id = None
        self.DBID = None
        self.cDNA = None
        self.gDNA = None
        self.mRNA = None
        self.times_reported = None
        self.link = None
        self.alt_link = None
        self.references = {'pmid': [], 'doi': []}

        if vdict is not None:
            self.__dict__.update(vdict)
        elif table_row is not None:
            self.__dict__.update(_get_vdict_from_tr(table_row, has_haplotype))
            if transcript:
                self.hgvs_text = construct_hgvs_name(transcript, self.cDNA)

    def load_from_api(self):
        """ Loads information from LOVD variant API, which does not return certain items like
        dbSNP number or references."""
        pass

    def load_from_page(self):
        """ Loads information from LOVD variant page, which is slower than an API call (and will
        probably break more readily over time), but includes references and other useful info."""
        pass

    def to_dict(self):
        return self.__dict__    


def _get_references_from_td(td_elem):
    refs = {'doi': [], 'pmid': []}
    for span in td_elem.findall('span'):
        if not span.text:
            continue

        if span.text.startswith('Journal'):
            stuff = span.get('onmouseover')
            doi = re_doi.findall(stuff)[0].strip('\\')
            if doi:
                refs['doi'].append(doi)
            
        if span.text.startswith('PubMed'):
            if span.text.startswith('PubMed'):
                stuff = span.get('onmouseover')
                pmid = re_pubmed_id.findall(stuff)[0]
                if pmid:
                    refs['pmid'].append(pmid)
    return refs


def _get_text_from_link_in_td(td_elem):
    try:
        return td_elem.find('a').text
    except AttributeError:
        return ''


def _get_lovd_link_from_td(td_elem):
    return td_elem.find('a').get('href')


def _get_vdict_from_tr(tr_elem, has_haplotype=False):
    tds = tr_elem.getchildren()

    # there are 2 different styles of page. one has Haplotype at td-5, the other has no Haplotype column.
    # we have to parse these out differently.

    if has_haplotype:
        return { 'effect': tds[0].text,
                 'exon': tds[1].text,
                 'cDNA': _get_text_from_link_in_td(tds[2]), 
                 'lovd_link': _get_lovd_link_from_td(tds[2]),
                 'mRNA': tds[3].text,
                 'protein': tds[4].text,
                 'haplotype': tds[5].text,
                 'allele': tds[6].text,
                 'gDNA': _get_text_from_link_in_td(tds[7]),
                 'DBID': tds[10].text,
                 'references': _get_references_from_td(tds[12]),
                 'frequency': tds[16].text,
                 'remarks': tds[25].text,
                }
    else:
        return { 'effect': tds[0].text,
                 'exon': tds[1].text,
                 'cDNA': _get_text_from_link_in_td(tds[3]), 
                 'lovd_link': _get_lovd_link_from_td(tds[2]),
                 'mRNA': tds[4].text,
                 'protein': tds[5].text,
                 'gDNA': _get_text_from_link_in_td(tds[6]),
                 'DBID': tds[9].text,
                 'remarks': tds[10].text,
                 'references': _get_references_from_td(tds[11]),
                 'dbSNP': tds[11].text,
                 'genetic_origin': tds[12].text,
                 'frequency': tds[14].text,
                }
         

def _parse_content_text(text):
    """ <content type="text">
          symbol:FANCA
          id:0000116087
          position_mRNA:c.1626+1_1627-1
          position_genomic:chr16:?
          Variant/DNA:c.(1626+1_1627-1)_(2151+1_2152-1)del
          Variant/DBID:FANCA_000723
          Times_reported:1
        </content>
    """
    out = {}
    for part in text.split('\n'):
        part = part.strip()
        try:
            key, val = part.split(':', 1)
        except IndexError:
            continue

        if part.startswith('symbol'):
            out['gene_name'] = val
        if part.startswith('id'):
            out['id'] = val
        if part.startswith('Variant/DBID'):
            out['DBID'] = val
        if part.startswith('Variant/DNA'):
            out['cDNA'] = val
        if part.startswith('position_mRNA'):
            out['mRNA'] = val
        if part.startswith('position_genomic'):
            out['gDNA'] = val
        if part.startswith('Times_reported'):
            out['times_reported'] = val
    return out


def _parse_entry(entry):
    vdict = _parse_content_text(entry['content']['#text'])
    vdict['link'] = entry['link'][1]['@href']
    vdict['alt_link'] = (entry['link'][0]['@href'])
    vdict['title'] = entry['title']
    vdict['transcript'] = re_transcript.findall(vdict['alt_link'])[0]
    vdict['hgvs_text'] = construct_hgvs_name(vdict['transcript'], entry['title'].replace(vdict['gene_name']+ ':', ''))
    return vdict


def _parse_lovd_variants_by_gene_name_response(xml_blob):
    """ Parses LOVD response as provided by query_lovd_for_variants_by_gene_name

    Returns list of dictionaries containing pared-down and interpreted variant info.

    :param xml_blob: (str) XML response from LOVD
    :return: list of dictionaries (one per variant entry)
    """
    entries = xmltodict.parse(xml_blob)['feed']['entry']
    out = []
    for entry in entries:
        out.append(_parse_entry(entry))
    return out


def _query_lovd_api_for_variants_by_gene_name(symbol, domain):
    """ Submits query to lovd.nl for specified gene name (Symbol), retrieving the API
    content for the given gene.

    :param symbol: (str)
    :param domain: (str) domain name of target LOVD database (default: databases.lovd.nl)
    :return: unparsed response content (str)
    :raises: LOVDRemoteError if not response.ok (message contains response.status_code)
    """

    response = requests.get(LOVD_API_GENE_VARIANTS_URL.format(symbol=symbol, domain=domain))
    if response.ok:
        return response.content
    else:
        raise LOVDRemoteError('lovd.nl returned HTTP %r' % response.status_code)


def _load_lovd_variant_page_by_gene_name(symbol, domain):
    """ Submits query to lovd.nl for specified gene name (Symbol), retrieving the WEBSITE
    content for the given gene.

    :param symbol: (str)
    :param domain: (str) domain name of target LOVD database (default: databases.lovd.nl)
    :return: unparsed response content (str)
    :raises: LOVDRemoteError if not response.ok (message contains response.status_code)
    """
    response = requests.get(LOVD_PAGE_GENE_VARIANTS_URL.format(symbol=symbol, domain=domain))
    if response.ok:
        return response.content
    else:
        raise LOVDRemoteError('lovd.nl returned HTTP %r' % response.status_code)


def get_variants_for_gene_name(symbol, domain=LOVD_DEFAULT_DOMAIN):
    """ Given HUGO gene name symbol (e.g. "ACVRL1", "FANCA"), return the set of unique variants 
        found on LOVD.
    
    NOTE: Not all genes can be found on lovd.nl -- see http://databases.lovd.nl/shared/genes

    :param symbol: (str)
    :param domain: (str) [default: lovd.nl]
    :return: variants (list) 
    """
    api_response = _query_lovd_api_for_variants_by_gene_name(symbol, domain)
    vdicts = _parse_lovd_variants_by_gene_name_response(api_response)
    lovd_variants = set([vdict['hgvs_text'] for vdict in vdicts])
    return lovd_variants


def get_variants_with_annotations_for_gene_name(symbol, domain=LOVD_DEFAULT_DOMAIN):
    """ Given HUGO gene name symbol (e.g. "ACVRL1", "FANCA"), return the set of unique variants 
        found on LOVD paired with references (citations, annotations, etc).
    
    NOTE: Not all genes can be found on lovd.nl -- see http://databases.lovd.nl/shared/genes

    :param symbol: (str)
    :param domain: (str) [default: lovd.nl]
    :return: variants (list) 
    """
    content = _load_lovd_variant_page_by_gene_name(symbol, domain)
    htm = etree.fromstring(content, parser=HTMLParser()).find('body')
    try:
        table = htm.cssselect('#viewlistTable_CustomVL_VIEW_%s' % symbol)[0]
    except:
        table = htm.cssselect('#viewlistTable_CustomVL_VOTunique_VOG_%s' % symbol)[0]
    transcripts = re.findall('using the (?P<acc_id>NM_\d+\.\d+) transcript reference sequence.', content)
    if transcripts:
        tran = transcripts[0]
    else:
        tran = None 

    # determine which type of table this is by what the headers look like
    has_haplotype = False
    ths = table.find('thead').getchildren()[0].findall('th')
    for th in ths:
        title = th.get('title')
        if title and 'haplotype' in th.get('title').lower():
            has_haplotype = True
    return [LOVDVariant(table_row=tr, transcript=tran, has_haplotype=has_haplotype) for tr in table.findall('tr')]


LOVDVariantsForGene = get_variants_for_gene_name

LOVDAnnotatedVariants = get_variants_with_annotations_for_gene_name




import glob, os

from setuptools import setup, find_packages
from setuptools.command.build_ext import build_ext as _build_ext


class build_ext(_build_ext):
    # http://stackoverflow.com/questions/21605927/why-doesnt-setup-requires-work-properly-for-numpy/21621493
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

setup (
    name = 'metavariant',
    version = '0.1.2',
    description = 'a lexical manipulation toolkit for genetic variant descriptors (hgvs, etc)',
    long_description = open('README.rst').read(),
    long_description_content_type = 'text/x-rst',
    author = 'Naomi Most',
    maintainer = 'Naomi Most',
    author_email = 'naomi@nthmost.com',
    maintainer_email = 'naomi@nthmost.com',
    license = 'Apache 2.0',
    packages = find_packages(),
    cmdclass = {'build_ext': build_ext},
    setup_requires = ['setuptools', 'numpy'],
    install_requires = [
        'numpy',
        #'uta',     # let this be installed by hgvs
        'hgvs>=1.3.0.post0',
        'xmltodict',
        'pytest',
        ],
    )

